#include "cudaConvolution.h"

#include <math.h>
#include <stdio.h>
#include <string.h>

// We will use 1D blocks in our solution, so each block has ones of
// these values in either x or y and size 1 in the other.  It is
// assumed that the block size is larger than the stencil size.
#define BLOCK_SIZE_X 256
#define BLOCK_SIZE_Y 256

// Size of the stencils.  Do not modify.
#define STENCIL_WIDTH_X 21
#define STENCIL_WIDTH_Y 11

// Global variables to store the convolution stencils.
float *hos_stencil_1dx = NULL;
float *hos_stencil_1dy = NULL;

// Declaration to create pointers to GPU constant memory.
__constant__ float dev_stencil_1dx[STENCIL_WIDTH_X];
__constant__ float dev_stencil_1dy[STENCIL_WIDTH_Y];

////////////////////////////////////////////////////////////////
///////////////////////// CUDA kernels /////////////////////////
////////////////////////////////////////////////////////////////

__global__ void conv1h_basic_kernel(int width, int height,
				    float *dev_input, float *dev_output) {
  int x = blockDim.x * blockIdx.x + threadIdx.x;
  int y = blockIdx.y;

  if (x >= width)
    return;

  int stencil_half_width = STENCIL_WIDTH_X / 2;

  // Compute 1D convolution in parallel
  float sum = 0.f;
  for (int i = 0; i < STENCIL_WIDTH_X; i++) {
    int cur_x = x + i - stencil_half_width;
    int image_offset = y * width + cur_x;
    if (cur_x >= 0 && cur_x < width) {
      sum += dev_stencil_1dx[i] * dev_input[image_offset];
    }
  }

  int image_offset = y * width + x;
  dev_output[image_offset] = sum;
}

__global__ void conv1v_basic_kernel(int width, int height,
				    float *dev_input, float *dev_output) {
  // The same as horizontal convolution except that we use the x and y
  // dimensions of the block in the opposite sense (eg: threads in x
  // dimension work on the vertical part of the image, threads in y
  // dimension work on horizontal part of the image.
  int y = blockDim.x * blockIdx.x + threadIdx.x;
  int x = blockIdx.y;

  if (y >= height)
    return;

  int stencil_half_width = STENCIL_WIDTH_Y / 2;

  // Compute 1D convolution in parallel
  float sum = 0.f;
  for (int i = 0; i < STENCIL_WIDTH_Y; i++) {
    int cur_y = y + i - stencil_half_width;
    int image_offset = cur_y * width + x;
    if (cur_y >= 0 && cur_y < height)
      sum += dev_stencil_1dy[i] * dev_input[image_offset];
  }

  int image_offset = y * width + x;
  dev_output[image_offset] = sum;
}

__global__ void conv1h_tiled_kernel(int width, int height,
				    float *dev_input, float *dev_output) {
  int x = blockDim.x * blockIdx.x + threadIdx.x;
  int y = blockIdx.y;

  // Move to shared memory
  __shared__ float s_inputs[STENCIL_WIDTH_X - 1 + BLOCK_SIZE_X];

  // Since the standard answer is in the book
  // I'm going to use a different (I think simpler) method

  // First get thread offset within the block (0, ..., BLOCK_SIZE)
  int t = threadIdx.x;

  // Then get the target range of x (left halo + tile + right halo)
  int stencil_half_width = STENCIL_WIDTH_X / 2;
  int halo_left_start = blockDim.x * blockIdx.x - stencil_half_width;
  // halo_right_end is the index after the last right halo
  int halo_right_end = blockDim.x * (blockIdx.x + 1) + stencil_half_width;

  // Move the block:
  // Ex: "Round" 0 (not really a round since there's no synchronization)
  // | block | ->
  // | left, tile, right |
  // "Round" 1
  //          | block | ->
  // | left, tile, right |
  for (int i = t; i + halo_left_start < halo_right_end; i += BLOCK_SIZE_X) {
    int cur_x = i + halo_left_start;
    int image_offset = y * width + cur_x;
    if (i + halo_left_start < width && 
      i + halo_left_start >= 0)
      s_inputs[i] = dev_input[image_offset];
    else
      s_inputs[i] = 0.f;
  }

  __syncthreads();

  // Compute 1D convolution in parallel
  float sum = 0.f;
  for (int i = 0; i < STENCIL_WIDTH_X; i++) {
    int cur_x = t + i;
    sum += dev_stencil_1dx[i] * s_inputs[cur_x];
  }

  int image_offset = y * width + x;
  if (x < width)
    dev_output[image_offset] = sum;
}

__global__ void conv1v_tiled_kernel(int width, int height,
				    float *dev_input, float *dev_output) {
  // The same as horizontal convolution except that we use the x and y
  // dimensions of the block in the opposite sense (eg: threads in x
  // dimension work on the vertical part of the image, threads in y
  // dimension work on horizontal part of the image.

  int y = blockDim.x * blockIdx.x + threadIdx.x;
  int x = blockIdx.y;

  // Move to shared memory
  __shared__ float s_inputs[STENCIL_WIDTH_Y - 1 + BLOCK_SIZE_Y];

  // Since the standard answer is in the book
  // I'm going to use a different (I think simpler) method

  // First get thread offset within the block (0, ..., BLOCK_SIZE)
  int t = threadIdx.x;

  // Then get the target range of x (left halo + tile + right halo)
  int stencil_half_width = STENCIL_WIDTH_Y / 2;
  int halo_left_start = blockDim.x * blockIdx.x - stencil_half_width;
  // halo_right_end is the index after the last right halo
  int halo_right_end = blockDim.x * (blockIdx.x + 1) + stencil_half_width;

  // Move the block:
  // Ex: "Round" 0 (not really a round since there's no synchronization)
  // | block | ->
  // | left, tile, right |
  // "Round" 1
  //          | block | ->
  // | left, tile, right |
  for (int i = t; i + halo_left_start < halo_right_end; i += BLOCK_SIZE_Y) {
    int cur_y = i + halo_left_start;
    int image_offset = cur_y * width + x;
    if (i + halo_left_start < height && 
      i + halo_left_start >= 0)
      s_inputs[i] = dev_input[image_offset];
    else
      s_inputs[i] = 0.f;
  }

  __syncthreads();

  // Compute 1D convolution in parallel
  float sum = 0.f;
  for (int i = 0; i < STENCIL_WIDTH_Y; i++) {
    int cur_y = t + i;
    sum += dev_stencil_1dy[i] * s_inputs[cur_y];
  }

  int image_offset = y * width + x;
  if (y < height)
    dev_output[image_offset] = sum;
}

//////////////////////////////////////////////////////////////////
///////////////////////// Host functions /////////////////////////
//////////////////////////////////////////////////////////////////
// TO DO: Modify the code in the kernels below to answer the homework
// questions.
//
// Notes:
//
// float *hos_stencil_1dx is a host global pointer containing a 1D
// array of length STENCIL_WIDTH_X with the stencil data to be used for the
// horizontal convolution.
//
// float *hos_stencil_1dy is a host global pointer containing a 1D
// array of length STENCIL_WIDTH_Y with the stencil data to be used for the
// vertical convolution.

void conv1h_cpu(int width, int height, 
		float *hos_data_in, float *hos_data_out) {

  int stencil_half_width = STENCIL_WIDTH_X / 2;
  for(int i = 0; i < height; i++)
    for(int j = 0; j < width; j++) {
      float sum = 0.0;
      for(int k = 0; k < STENCIL_WIDTH_X; k++) {
	int image_index = j + k - stencil_half_width;
	if((image_index >= 0) && (image_index < width))
	  sum += hos_stencil_1dx[k] * hos_data_in[i * width + image_index];
      }
      hos_data_out[i * width + j] = sum;
    }
}


void conv1h_basic(int width, int height, 
		  float *hos_data_in, float *hos_data_out) {
  float *dev_image_in_buffer;
  float *dev_image_out_buffer;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_in_buffer, image_size);
  cudaMalloc(&dev_image_out_buffer, image_size);

  cudaMemcpy(dev_image_in_buffer, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);

  // Copy stencil to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dx, hos_stencil_1dx, 
		     sizeof(float) * STENCIL_WIDTH_X);

  // Compute grid and block size
  dim3 blockDim(BLOCK_SIZE_X, 1, 1);
  dim3 gridDim((width + blockDim.x - 1) / blockDim.x, height, 1);

  // Launch kernel
  conv1h_basic_kernel<<<gridDim, blockDim>>>(width, height,
    dev_image_in_buffer, dev_image_out_buffer);

  // Retrieve the output image and free the memory on the device.
  cudaMemcpy(hos_data_out, dev_image_out_buffer, image_size,
	     cudaMemcpyDeviceToHost);

  cudaFree(dev_image_in_buffer);
  cudaFree(dev_image_out_buffer);
}

// Q2 (b)
void conv1v_basic(int width, int height, float *hos_data_in, float *hos_data_out) {
  float *dev_image_in_buffer;
  float *dev_image_out_buffer;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_in_buffer, image_size);
  cudaMalloc(&dev_image_out_buffer, image_size);

  cudaMemcpy(dev_image_in_buffer, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);

  // Copy stencil to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dy, hos_stencil_1dy, 
		     sizeof(float) * STENCIL_WIDTH_Y);

  // Compute grid and block size.  Keep in mind that the kernel will
  // swap the x and y dimensions.
  dim3 blockDim(BLOCK_SIZE_Y, 1, 1);
  dim3 gridDim((height + blockDim.x - 1) / blockDim.x, width, 1);

  // Launch kernel
  conv1v_basic_kernel<<<gridDim, blockDim>>>(width, height,
    dev_image_in_buffer, dev_image_out_buffer);

  // Retrieve the output image and free the memory on the device.
  cudaMemcpy(hos_data_out, dev_image_out_buffer, image_size,
	     cudaMemcpyDeviceToHost);

  cudaFree(dev_image_in_buffer);
  cudaFree(dev_image_out_buffer);
}

// Q2 (c)
void conv1to2_basic(int width, int height, float *hos_data_in, float *hos_data_out) {
  float *dev_image_buffer_1;
  float *dev_image_buffer_2;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_buffer_1, image_size);
  cudaMalloc(&dev_image_buffer_2, image_size);

  cudaMemcpy(dev_image_buffer_1, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);

  // Copy stencils to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dx, hos_stencil_1dx, 
		     sizeof(float) * STENCIL_WIDTH_X);
  cudaMemcpyToSymbol(dev_stencil_1dy, hos_stencil_1dy, 
		     sizeof(float) * STENCIL_WIDTH_Y);

  // Compute grid and block size
  dim3 blockDim_h(BLOCK_SIZE_X, 1, 1);
  dim3 blockDim_v(BLOCK_SIZE_Y, 1, 1);
  dim3 gridDim_h((width + blockDim_v.x - 1) / blockDim_h.x, height, 1);
  dim3 gridDim_v((height + blockDim_h.x - 1) / blockDim_v.x, width, 1);

  // Launch kernels sequentially.
  // First launch uses input image in buffer_1 and output to buffer_2.
  conv1h_basic_kernel<<<gridDim_h, blockDim_h>>>(width, height,
    dev_image_buffer_1, dev_image_buffer_2);

  // Second launch uses input image in buffer_2 and output to buffer_1.
  conv1v_basic_kernel<<<gridDim_v, blockDim_v>>>(width, height,
    dev_image_buffer_2, dev_image_buffer_1);

  // Retrieve the results from buffer_1.
  cudaMemcpy(hos_data_out, dev_image_buffer_1,
         width * height * sizeof(float),
         cudaMemcpyDeviceToHost);

  cudaFree(dev_image_buffer_1);
  cudaFree(dev_image_buffer_2);
}

// These host routines are the same as those above, except for calling
// the tiled kernels.
void conv1h_tiled(int width, int height, 
		  float *hos_data_in, float *hos_data_out) {
  float *dev_image_in_buffer;
  float *dev_image_out_buffer;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_in_buffer, image_size);
  cudaMalloc(&dev_image_out_buffer, image_size);

  cudaMemcpy(dev_image_in_buffer, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);


  // Copy stencil to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dx, hos_stencil_1dx, 
		     sizeof(float) * STENCIL_WIDTH_X);

  // Compute grid and block size
  dim3 blockDim(BLOCK_SIZE_X, 1, 1);
  dim3 gridDim((width + blockDim.x - 1) / blockDim.x, height, 1);

  // Launch kernel
  conv1h_tiled_kernel<<<gridDim, blockDim>>>(width, height,
    dev_image_in_buffer, dev_image_out_buffer);

  // Retrieve the output image and free the memory on the device.
  cudaMemcpy(hos_data_out, dev_image_out_buffer, image_size,
	     cudaMemcpyDeviceToHost);

  cudaFree(dev_image_in_buffer);
  cudaFree(dev_image_out_buffer);
}

void conv1v_tiled(int width, int height, 
		  float *hos_data_in, float *hos_data_out) {
  float *dev_image_in_buffer;
  float *dev_image_out_buffer;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_in_buffer, image_size);
  cudaMalloc(&dev_image_out_buffer, image_size);

  cudaMemcpy(dev_image_in_buffer, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);

  // Copy stencil to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dy, hos_stencil_1dy, 
		     sizeof(float) * STENCIL_WIDTH_Y);

  // Compute grid and block size.  Keep in mind that the kernel will
  // swap the x and y dimensions.
  dim3 blockDim(BLOCK_SIZE_Y, 1, 1);
  dim3 gridDim((height + blockDim.x - 1) / blockDim.x, width, 1);

  // Launch kernel
  conv1v_tiled_kernel<<<gridDim, blockDim>>>(width, height,
    dev_image_in_buffer, dev_image_out_buffer);

  // Retrieve the output image and free the memory on the device.
  cudaMemcpy(hos_data_out, dev_image_out_buffer, image_size,
	     cudaMemcpyDeviceToHost);

  cudaFree(dev_image_in_buffer);
  cudaFree(dev_image_out_buffer);
}

void conv1to2_tiled(int width, int height, 
		    float *hos_data_in, float *hos_data_out) {
  float *dev_image_buffer_1;
  float *dev_image_buffer_2;

  // Allocate space on the device and copy over the input image.
  int image_size = width * height * sizeof(float);

  cudaMalloc(&dev_image_buffer_1, image_size);
  cudaMalloc(&dev_image_buffer_2, image_size);

  cudaMemcpy(dev_image_buffer_1, hos_data_in, image_size,
	     cudaMemcpyHostToDevice);

  // Copy stencils to constant memory
  cudaMemcpyToSymbol(dev_stencil_1dx, hos_stencil_1dx, 
		     sizeof(float) * STENCIL_WIDTH_X);
  cudaMemcpyToSymbol(dev_stencil_1dy, hos_stencil_1dy, 
		     sizeof(float) * STENCIL_WIDTH_Y);

  // Compute grid and block size
  dim3 blockDim_h(BLOCK_SIZE_X, 1, 1);
  dim3 blockDim_v(BLOCK_SIZE_Y, 1, 1);
  dim3 gridDim_h((width + blockDim_v.x - 1) / blockDim_h.x, height, 1);
  dim3 gridDim_v((height + blockDim_h.x - 1) / blockDim_v.x, width, 1);

  // Launch kernels sequentially.
  // First launch uses input image in buffer_1 and output to buffer_2.
  conv1h_tiled_kernel<<<gridDim_h, blockDim_h>>>(width, height,
    dev_image_buffer_1, dev_image_buffer_2);

  // Second launch uses input image in buffer_2 and output to buffer_1.
  conv1v_tiled_kernel<<<gridDim_v, blockDim_v>>>(width, height,
    dev_image_buffer_2, dev_image_buffer_1);

  // Retrieve the results from buffer_1.
  cudaMemcpy(hos_data_out, dev_image_buffer_1,
         width * height * sizeof(float),
         cudaMemcpyDeviceToHost);

  cudaFree(dev_image_buffer_1);
  cudaFree(dev_image_buffer_2);
}

/////////////////////////////////////////////////////////////////////////
///////////////// No change to code after this point ////////////////////
/////////////////////////////////////////////////////////////////////////
// DO NOT CHANGE THE CODE BELOW THIS COMMENT (or in any of the other
// files).  Modification of the code below or in the other files may
// cause the autograder to fail, and you may receive a zero for the
// corresponding questions in the homework.

Convolution::~Convolution() {
  free(hos_stencil_1dx);
  free(hos_stencil_1dy);
  
  hos_stencil_1dx = hos_stencil_1dy = NULL;
}

void Convolution::evaluate_gaussian_stencil(float sigma, float *stencil_array,
					    int stencil_width) {
  // Compute the stencil
  float normalization = 0.0f;
  int half_width = stencil_width / 2;
  float interval = 20.0f / stencil_width;
  for (int i = 0; i < stencil_width; i++) {
    float x = (i - half_width) * interval;
    float gaussian = std::exp(-(x * x) / (2 * sigma *sigma));
    stencil_array[i] = gaussian;

    normalization += gaussian;
    //printf("%d: %f - %f\n", i, x, gaussian);
  }

  // Normalize so that stencil sums to 1 and store to stencil_array.
  for (int i = 0; i < stencil_width; i++)
    stencil_array[i] /= normalization;
}


void Convolution::setup_host_gaussian_stencil(float sigma) {
  // Allocate memory, freed in destructor.
  hos_stencil_1dx = (float*)malloc(STENCIL_WIDTH_X * sizeof(float));
  hos_stencil_1dy = (float*)malloc(STENCIL_WIDTH_Y * sizeof(float));

  // Evaluate Gaussian function to create the stencils.
  evaluate_gaussian_stencil(sigma, hos_stencil_1dx, STENCIL_WIDTH_X);
  evaluate_gaussian_stencil(sigma, hos_stencil_1dy, STENCIL_WIDTH_Y);
}

void Convolution::setup_device(int width, int height) {
  cudaDeviceProp prop;

  int ndev;
  cudaGetDeviceCount(&ndev);
  if(ndev < 1) {
    fprintf(stderr, "No CUDA device found\n");
    exit(-1);
  }
  cudaGetDeviceProperties(&prop, 0);

  printf("The GPU is a %s\n", prop.name);
  printf("Cuda capability %d.%d.\n", prop.major, prop.minor);
  printf("Shared memory per block %d bytes.\n", prop.sharedMemPerBlock);
}

void Convolution::load_image_input(const Image &image) {
  if (image.channel != 1) {
    printf("Error: Input image has %d channels (should be 1).\n", image.channel);
  }

  int w = image.width, h = image.height;

  // Allocate host input image buffer
  if (!hos_image_in.pixels) {
    hos_image_in.alloc(image.width, image.height, image.channel);
  }

  memcpy(hos_image_in.pixels, image.pixels, w * h * sizeof(float));
  hos_image_out.alloc(w, h, 1);
}

// Since we've copied input to device in function load_image_input,
// we can just launch kernels here:
void Convolution::run_horizontal_1d_cpu() {
  int width = hos_image_in.width, height = hos_image_in.height;

  // Call student's code
  conv1h_cpu(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_horizontal_1d() {
  int width = hos_image_in.width, height = hos_image_in.height;

  // Call student's code
  conv1h_basic(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_vertical_1d() {
  int width = hos_image_in.width, height = hos_image_in.height;

  // Call student's code
  conv1v_basic(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_1to2() {
  int width = hos_image_in.width, height = hos_image_in.height;
  
  conv1to2_basic(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_1to2_tiling() {
  int width = hos_image_in.width, height = hos_image_in.height;

  conv1to2_tiled(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_2d_tiling() {

}

void Convolution::run_horizontal_1d_tiling() {
  int width = hos_image_in.width, height = hos_image_in.height;

  // Call student's code
  conv1h_tiled(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

void Convolution::run_vertical_1d_tiling() {
  int width = hos_image_in.width, height = hos_image_in.height;

    // Call student's code
  conv1v_tiled(width, height, hos_image_in.pixels, hos_image_out.pixels);
}

const float *Convolution::get_image_buffer() {
  return hos_image_out.pixels;
}

const Image *Convolution::get_image_ptr() {
  return &hos_image_out;
}
